@@ -15,6 +15,7 @@ from bam_masterdata.checker import MasterdataChecker
 from bam_masterdata.cli.entities_to_excel import entities_to_excel
 from bam_masterdata.cli.entities_to_rdf import entities_to_rdf
 from bam_masterdata.cli.fill_masterdata import MasterdataCodeGenerator
+from bam_masterdata.cli.run_parser import run_parser
 from bam_masterdata.logger import logger
 from bam_masterdata.metadata.entities_dict import EntitiesDict
 from bam_masterdata.openbis.login import ologin
@@ -56,6 +57,7 @@ def cli():
 @click.option(
     "--export-dir",
     type=str,
+    default="./bam_masterdata/datamodel",
     required=False,
     help="The directory where the Masterdata will be exported to.",
 )
@@ -581,5 +583,61 @@ def push_to_openbis(file_path, datamodel_path):
     )
 
 
+@cli.command(
+    name="parser",
+    help="Parses a list of files using the specified parsers and stores the results in openBIS.",
+)
+@click.option(
+    "--files-parser",
+    "files_parser",  # alias
+    multiple=True,
+    type=click.Tuple(
+        [str, click.Path()],
+    ),
+    help="Parser name and file path tuple: 'ExampleParser file1.txt'",
+)
+@click.option(
+    "--project-name",
+    "project_name",  # alias
+    type=str,
+    required=True,
+    help="OpenBIS project name",
+)
+@click.option(
+    "--collection-name",
+    "collection_name",  # alias
+    type=str,
+    required=False,
+    help="OpenBIS collection name",
+)
+@click.option(
+    "--space-name",
+    "space_name",  # alias
+    type=str,
+    required=False,
+    help="OpenBIS space name",
+)
+def parser(files_parser, project_name, collection_name, space_name):
+    parser_map = {}  # TODO load from configuration from yaml file
+    parse_file_dict = {}
+    for parser_key, filepath in files_parser:
+        if parser_key not in parser_map:
+            logger.warning(
+                f"Parser {parser_key} not found. Available parsers are: {', '.join(parser_map.keys())}"
+            )
+
+            continue
+        parser_cls = parser_map[parser_key]
+        parse_file_dict[parser_cls].append(filepath)
+
+    run_parser(
+        openbis=ologin(url=environ("OPENBIS_URL")),
+        space_name=space_name,
+        project_name=project_name,
+        collection_name=collection_name,
+        files_parser=parse_file_dict,
+    )
+
+
 if __name__ == "__main__":
     cli()
