@@ -160,6 +160,8 @@ class MasterdataExcelExtractor:
                     sheet_title=sheet_title,
                 )
         elif is_data:
+            # Normalize data type to uppercase
+            val = val.upper()
             if val not in [dt.value for dt in DataType]:
                 self.logger.error(
                     error_message
@@ -169,7 +171,6 @@ class MasterdataExcelExtractor:
                     cell_coordinate=coordinate,
                     sheet_title=sheet_title,
                 )
-                val = val.upper()
         elif is_url:
             if not re.match(
                 r"https?://(?:www\.)?[a-zA-Z0-9-._~:/?#@!$&'()*+,;=%]+", val
@@ -541,7 +542,7 @@ class MasterdataExcelExtractor:
         extracted_columns: dict[str, list] = {term: [] for term in expected_terms}
 
         # Helper function to process each term
-        def process_term(term, cell_value, coordinate, sheet_title):
+        def process_term_cell(term, cell_value, coordinate, sheet_title):
             if term == "Official":
                 return self.str_to_bool(
                     value=cell_value,
@@ -573,7 +574,7 @@ class MasterdataExcelExtractor:
             # Extract values from the column
             for cell in sheet[term_letter][header_index:last_non_empty_row]:
                 extracted_columns[term].append(
-                    process_term(term, cell.value, cell.coordinate, sheet.title)
+                    process_term_cell(term, cell.value, cell.coordinate, sheet.title)
                 )
 
         # Combine extracted values into a dictionary
@@ -581,11 +582,16 @@ class MasterdataExcelExtractor:
             terms_dict[extracted_columns["Code"][i]] = {
                 "permId": extracted_columns["Code"][i],
                 "code": extracted_columns["Code"][i],
-                "descriptions": extracted_columns["Description"][i],
-                "url_template": extracted_columns["Url template"][i],
-                "label": extracted_columns["Label"][i],
-                "official": extracted_columns["Official"][i],
             }
+            for key, pybis_val in {
+                "Description": "descriptions",
+                "Url template": "url_template",
+                "Label": "label",
+                "Official": "official",
+            }.items():
+                if extracted_columns.get(key):
+                    value = extracted_columns[key][i]
+                    terms_dict[extracted_columns["Code"][i]][pybis_val] = value
 
         return terms_dict
 
@@ -645,11 +651,8 @@ class MasterdataExcelExtractor:
         """
         Extracts entities from an Excel file and returns them as a dictionary.
 
-        Args:
-            output_directory: The directory to store the output files.
-
         Returns:
-            A dictionary where each key is a normalized sheet name and the value is a dictionary
+            dict[str, dict[str, Any]]: A dictionary where each key is a normalized sheet name and the value is a dictionary
             containing the extracted entities. Returns an empty dictionary if all sheets are empty.
         """
         sheets_dict: dict[str, dict[str, Any]] = {}
