@@ -1,3 +1,5 @@
+import datetime
+import inspect
 import json
 from collections.abc import Callable
 from typing import TYPE_CHECKING, Any, no_type_check
@@ -7,6 +9,8 @@ from pydantic import BaseModel, ConfigDict, Field, model_validator
 from rdflib import BNode, Literal
 from rdflib.namespace import DC, OWL, RDF, RDFS
 
+from bam_masterdata.utils import DATAMODEL_DIR, import_module, listdir_py_modules
+
 if TYPE_CHECKING:
     from pybis import Openbis
     from rdflib import Graph, Namespace, URIRef
@@ -38,6 +42,13 @@ class BaseEntity(BaseModel):
     adding new methods that are useful for interfacing with openBIS.
     """
 
+    code: str | None = Field(
+        default=None,
+        description="""
+        Code of the entity to assign as permanent identifier in openBIS.
+        """,
+    )
+
     def __init__(self, **kwargs):
         super().__init__()
 
@@ -66,18 +77,26 @@ class BaseEntity(BaseModel):
 
     def __repr__(self):
         # Filter for attributes that are `PropertyTypeAssignment` and set to a finite value
+        class_prop_name = None
         fields = []
         for key, metadata in self._property_metadata.items():
             if isinstance(metadata, PropertyTypeAssignment):
                 value = getattr(self, key, None)
                 # Only include set attributes
                 if value is not None and not isinstance(value, PropertyTypeAssignment):
+                    if key == "name":
+                        class_prop_name = value
                     fields.append(f"{key}={repr(value)}")
 
         # Format the output
-        class_name = self.__class__.__name__
+        class_name = self.cls_name
+        if class_prop_name:  # adding `name` if available
+            class_name = f"{class_prop_name}:{class_name}"
         return f"{class_name}({', '.join(fields)})"
 
+    # Overwriting the __str__ method to use the same representation as __repr__
+    __str__ = __repr__
+
     @property
     def cls_name(self) -> str:
         """
@@ -263,9 +282,11 @@ class BaseEntity(BaseModel):
 
         # Store property metadata at class level
         prop_meta_dict: dict = {}
-        for attr_name, attr_value in cls_attrs.items():
-            if isinstance(attr_value, PropertyTypeAssignment):
-                prop_meta_dict[attr_name] = attr_value
+        for base in type(self).__mro__:
+            cls_attrs = getattr(base, "__dict__", {})
+            for attr_name, attr_value in cls_attrs.items():
+                if isinstance(attr_value, PropertyTypeAssignment):
+                    prop_meta_dict[attr_name] = attr_value
         return prop_meta_dict
 
     def to_json(self, indent: int | None = None) -> str:
@@ -282,9 +303,12 @@ class BaseEntity(BaseModel):
         data: dict = {}
         for key in self._property_metadata.keys():
             try:
-                data[key] = getattr(self, key)
+                value = getattr(self, key)
             except AttributeError:
                 continue
+            if isinstance(value, PropertyTypeAssignment):
+                continue
+            data[key] = value
         return json.dumps(data, indent=indent)
 
     def to_dict(self) -> dict:
@@ -485,6 +509,77 @@ class BaseEntity(BaseModel):
             graph.add((entity_uri, RDFS.subClassOf, restriction))
 
 
+class VocabularyType(BaseEntity):
+    """
+    Base class used to define vocabulary types. All vocabulary types must inherit from this class. The
+    vocabulary types are defined in the module `bam_masterdata/vocabulary_types.py`.
+
+    The `VocabularyType` class contains a list of all `terms` defined for a `VocabularyType`, for
+    internally represent the model in other formats (e.g., JSON or Excel).
+    """
+
+    model_config = ConfigDict(ignored_types=(VocabularyTypeDef, VocabularyTerm))
+
+    terms: list[VocabularyTerm] = Field(
+        default=[],
+        description="""
+        List of vocabulary terms. This is useful for internal representation of the model.
+        """,
+    )
+
+    @property
+    def base_name(self) -> str:
+        """
+        Returns the entity name of the class as a string.
+        """
+        return "VocabularyType"
+
+    @model_validator(mode="after")
+    @classmethod
+    def model_validator_after_init(cls, data: Any) -> Any:
+        """
+        Validate the model after instantiation of the class.
+
+        Args:
+            data (Any): The data containing the fields values to validate.
+
+        Returns:
+            Any: The data with the validated fields.
+        """
+        # Add all the vocabulary terms defined in the vocabulary type to the `terms` list.
+        # TODO check if the order is properly assigned
+        for base in cls.__mro__:
+            for attr_name, attr_val in base.__dict__.items():
+                if isinstance(attr_val, VocabularyTerm):
+                    data.terms.append(attr_val)
+
+        return data
+
+    def to_openbis(
+        self,
+        logger: "BoundLoggerLazyProxy",
+        openbis: "Openbis",
+        type: str = "vocabulary",
+        type_map: dict = VOCABULARY_TYPE_MAP,
+    ) -> None:
+        def get_type(openbis: "Openbis", code: str):
+            return openbis.get_vocabulary(code)
+
+        def create_type(openbis: "Openbis", defs: VocabularyTypeDef, terms: list):
+            return openbis.new_vocabulary(
+                code=defs.code, description=defs.description, terms=terms
+            )
+
+        super()._to_openbis(
+            logger=logger,
+            openbis=openbis,
+            type=type,
+            type_map=type_map,
+            get_type=get_type,
+            create_type=create_type,
+        )
+
+
 class ObjectType(BaseEntity):
     """
     Base class used to define object types. All object types must inherit from this class. The
@@ -513,8 +608,163 @@ class ObjectType(BaseEntity):
         """,
     )
 
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+
+        # Initialize the properties list to store PropertyTypeAssignment instances
+        self._properties = {}
+        for key, prop in self._property_metadata.items():
+            self._properties[key] = prop.data_type
+
+    def _set_object_value(self, key, value):
+        """
+        Sets the value when the data type is OBJECT.
+        """
+        if isinstance(value, str):
+            # Validate the path format: /{space}/{project}/{collection}/{object} or /{space}/{project}/{object}
+            # If path is valid, store it as-is
+            if not value.startswith("/"):
+                raise ValueError(
+                    f"Invalid OBJECT path format for '{key}': Path must start with '/', got '{value}'"
+                )
+            path_parts = value.strip("/").split("/")
+            if len(path_parts) not in [3, 4]:
+                raise ValueError(
+                    f"Invalid OBJECT path format for '{key}': Expected '/<space>/<project>/<collection>/<object>' "
+                    f"or '/<space>/<project>/<object>', got '{value}'"
+                )
+            # * We don't validate if the object exists here as it requires pybis connection
+            # * That validation should be done when saving to openBIS
+        elif isinstance(value, ObjectType):
+            # Check if the object instance has a code
+            if not hasattr(value, "code") or value.code is None:
+                raise ValueError(
+                    f"Object instance for '{key}' must have a 'code' attribute set to be used as a reference"
+                )
+        else:
+            raise TypeError(
+                f"Invalid type for OBJECT property '{key}': Expected str (path) or ObjectType instance, "
+                f"got {type(value).__name__}"
+            )
+        return value
+
+    def _validate_controlled_vocabulary(self, meta, key, value) -> None:
+        """
+        Validates the value of a CONTROLLEDVOCABULARY.
+        """
+        vocabulary_code = meta[key].vocabulary_code
+        if not vocabulary_code:
+            raise ValueError(
+                f"Property '{key}' of type CONTROLLEDVOCABULARY must have a vocabulary_code defined."
+            )
+        vocab_path = None
+        for file in listdir_py_modules(DATAMODEL_DIR):
+            if "vocabulary_types.py" in file:
+                vocab_path = file
+                break
+        if vocab_path is None:
+            raise FileNotFoundError(
+                f"The file 'vocabulary_types.py' was not found in the directory specified by {DATAMODEL_DIR}."
+            )
+        vocabulary_class = self.get_vocabulary_class(vocabulary_code, vocab_path)
+        if vocabulary_class is None:
+            raise ValueError(
+                f"No matching vocabulary class found for vocabulary_code '{vocabulary_code}'."
+            )
+        codes = [term.code for term in vocabulary_class.terms]
+        if value not in codes:
+            raise ValueError(
+                f"{value} for {key} is not in the list of allowed terms for vocabulary."
+            )
+
+    def __setattr__(self, key, value):
+        if key in ["_property_metadata", "_properties", "code"]:
+            super().__setattr__(key, value)
+            return
+
+        # key search in every nested class
+        for base in type(self).__mro__:
+            prop_meta = getattr(base, "get_property_metadata", None)
+            if callable(prop_meta):
+                meta = (
+                    prop_meta(self)
+                    if base is not type(self)
+                    else self._property_metadata
+                )
+                if key in meta:
+                    # Type check
+                    expected_type = meta[key].data_type.pytype
+                    if expected_type is datetime.datetime:
+                        if isinstance(value, datetime.datetime):
+                            try:
+                                value = value.strftime(
+                                    "%Y-%m-%d %H:%M:%S"
+                                )  # create string
+                                expected_type = str
+                            except ValueError:
+                                raise ValueError(
+                                    f"Invalid datetime format for '{key}': Expected ISO format string, got '{value}'"
+                                )
+                        elif isinstance(value, str):
+                            try:
+                                datetime.datetime.fromisoformat(value)
+                                expected_type = str
+                            except ValueError:
+                                raise ValueError(
+                                    f"Invalid datetime format for '{key}': Expected ISO format string, got '{value}'"
+                                )
+                        else:
+                            raise TypeError(
+                                f"Invalid type for '{key}': Expected datetime or ISO format string, got {type(value).__name__}"
+                            )
+                    if expected_type and not isinstance(value, expected_type):
+                        raise TypeError(
+                            f"Invalid type for '{key}': Expected {expected_type.__name__}, got {type(value).__name__}"
+                        )
+
+                    # Get data type for additional checks
+                    data_type = meta[key].data_type
+                    # OBJECT check and attr assignment
+                    if data_type == "OBJECT":
+                        return object.__setattr__(
+                            self, key, self._set_object_value(key, value)
+                        )
+                    # CONTROLLEDVOCABULARY check
+                    if data_type == "CONTROLLEDVOCABULARY":
+                        self._validate_controlled_vocabulary(meta, key, value)
+
+                    # Setting attribute value after all checks
+                    return object.__setattr__(self, key, value)
+
+        raise KeyError(
+            f"Key '{key}' not found in any property_metadata of {type(self).__name__} or its bases."
+        )
+
+    def get_vocabulary_class(
+        self, vocabulary_code: str, vocab_path: str
+    ) -> VocabularyType | None:
+        """
+        Get the class instance of the vocabulary type defined by `vocabulary_code` in the Python module
+        specified by `vocab_path`.
+
+        Args:
+            vocabulary_code (str): Code of the vocabulary type to get.
+            vocab_path (str): Path to the module containing the vocabulary type definitions.
+
+        Returns:
+            VocabularyType | None: The class of the vocabulary type if found, otherwise None.
+        """
+        module = import_module(vocab_path)
+        vocabulary_class = None
+        for name, obj in inspect.getmembers(module, inspect.isclass):
+            if name == code_to_class_name(vocabulary_code):
+                vocabulary_class = obj()
+                break
+
+        return vocabulary_class
+
     @property
-    def cls_name(self) -> str:
+    def base_name(self) -> str:
         """
         Returns the entity name of the class as a string.
         """
@@ -535,7 +785,7 @@ class ObjectType(BaseEntity):
         # Add all the properties assigned to the object type to the `properties` list.
         # TODO check if the order is properly assigned
         for base in cls.__mro__:
-            for attr_name, attr_val in base.__dict__.items():
+            for _, attr_val in base.__dict__.items():
                 if isinstance(attr_val, PropertyTypeAssignment):
                     data.properties.append(attr_val)
 
@@ -609,77 +859,6 @@ def generate_object_relationship_id(parent_id: str, child_id: str) -> str:
     return f"{parent_id}>>{child_id}"
 
 
-class VocabularyType(BaseEntity):
-    """
-    Base class used to define vocabulary types. All vocabulary types must inherit from this class. The
-    vocabulary types are defined in the module `bam_masterdata/vocabulary_types.py`.
-
-    The `VocabularyType` class contains a list of all `terms` defined for a `VocabularyType`, for
-    internally represent the model in other formats (e.g., JSON or Excel).
-    """
-
-    model_config = ConfigDict(ignored_types=(VocabularyTypeDef, VocabularyTerm))
-
-    terms: list[VocabularyTerm] = Field(
-        default=[],
-        description="""
-        List of vocabulary terms. This is useful for internal representation of the model.
-        """,
-    )
-
-    @property
-    def cls_name(self) -> str:
-        """
-        Returns the entity name of the class as a string.
-        """
-        return "VocabularyType"
-
-    @model_validator(mode="after")
-    @classmethod
-    def model_validator_after_init(cls, data: Any) -> Any:
-        """
-        Validate the model after instantiation of the class.
-
-        Args:
-            data (Any): The data containing the fields values to validate.
-
-        Returns:
-            Any: The data with the validated fields.
-        """
-        # Add all the vocabulary terms defined in the vocabulary type to the `terms` list.
-        # TODO check if the order is properly assigned
-        for base in cls.__mro__:
-            for attr_name, attr_val in base.__dict__.items():
-                if isinstance(attr_val, VocabularyTerm):
-                    data.terms.append(attr_val)
-
-        return data
-
-    def to_openbis(
-        self,
-        logger: "BoundLoggerLazyProxy",
-        openbis: "Openbis",
-        type: str = "vocabulary",
-        type_map: dict = VOCABULARY_TYPE_MAP,
-    ) -> None:
-        def get_type(openbis: "Openbis", code: str):
-            return openbis.get_vocabulary(code)
-
-        def create_type(openbis: "Openbis", defs: VocabularyTypeDef, terms: list):
-            return openbis.new_vocabulary(
-                code=defs.code, description=defs.description, terms=terms
-            )
-
-        super()._to_openbis(
-            logger=logger,
-            openbis=openbis,
-            type=type,
-            type_map=type_map,
-            get_type=get_type,
-            create_type=create_type,
-        )
-
-
 class CollectionType(ObjectType):
     model_config = ConfigDict(
         ignored_types=(
@@ -709,8 +888,11 @@ class CollectionType(ObjectType):
         """,
     )
 
+    def __repr__(self):
+        return f"{self.base_name}(attached_objects={self.attached_objects}, relationships={self.relationships})"
+
     @property
-    def cls_name(self) -> str:
+    def base_name(self) -> str:
         """
         Returns the entity name of the class as a string.
         """
@@ -765,6 +947,19 @@ class CollectionType(ObjectType):
             raise TypeError(
                 f"Expected an ObjectType instance, got `{type(object_type).__name__}`"
             )
+
+        # Check mandatory properties are filled
+        missing_fields = []
+        for attr_name, prop in object_type._property_metadata.items():
+            assigned_prop = getattr(object_type, attr_name, None)
+            if prop.mandatory and isinstance(assigned_prop, PropertyTypeAssignment):
+                missing_fields.append(attr_name)
+
+        if missing_fields:
+            raise ValueError(
+                f"The following mandatory fields are missing for ObjectType '{object_type.cls_name}': {', '.join(missing_fields)}"
+            )
+
         object_id = generate_object_id(object_type)
         self.attached_objects[object_id] = object_type
         return object_id
@@ -833,7 +1028,7 @@ class CollectionType(ObjectType):
 
 class DatasetType(ObjectType):
     @property
-    def cls_name(self) -> str:
+    def base_name(self) -> str:
         """
         Returns the entity name of the class as a string.
         """
